#ifndef EXPIRATION_H_
#define EXPIRATION_H_

#include "config.h"
#include "meta.h"
#include "dirwalk.h"
#include "maintenance.h"
#include "lockable.h"
#include "csmapping.h"
#include "bgtask.h"
#include "tCacheProcessor.h"
#include "fileitem.h"
#include <set>

// #define USEDUPEFILTER

class dlcon;
class tDlJobHints;
struct foo;


// caching all relevant file identity data and helper flags in such entries
struct tDiskFileInfo
{
	time_t nLostAt;
	tFingerprint fpr;
	bool bHeaderTestDone :1;
	bool bForceContentOK :1;

	tDiskFileInfo() :
		nLostAt(0), bHeaderTestDone(false), bForceContentOK(false)
	{
	}

};

struct tFileNdir
{
	mstring file, dirRel;
	bool operator<(const tFileNdir &other) const
	{
		int nRel = file.compare(other.file);
		if (nRel)
			return nRel < 0;
		return dirRel.compare(other.dirRel) < 0;
	}
	inline tFileNdir(const mstring &f, const mstring &d) :
		file(f)
	{
		// help STL save some memory
		static mstring prev;
		if (prev == d)
			dirRel = prev;
		else
			prev = dirRel = d;
	}
};

typedef MYSTD::map<tFileNdir, tDiskFileInfo> tS2DAT;

struct tPatchEntry
{
	string patchName;
	tFingerprint fprState, fprPatch;
};
typedef deque<tPatchEntry>::const_iterator tPListConstIt;

typedef MYSTD::pair<tFingerprint,mstring> tContId;
struct tClassDesc {tStrDeq paths; tContId diffIdxId, bz2VersContId;};
typedef MYSTD::map<tContId, tClassDesc> tContId2eqClass;
typedef MYSTD::map<tContId, tClassDesc>::iterator tClassMapIter;

#define CACHE_BASE (acfg::cacheDirSlash)
#define CACHE_BASE_LEN (CACHE_BASE.length()) // where the relative paths begin

class tCacheMan :
	public IFileHandler,
	public tWuiBgTask,
	public ifileprocessor
{

public:
	tCacheMan(int);
	virtual ~tCacheMan();

protected:
	enum enumIndexType
	{
		EIDX_UNSUPPORTED =0,
		EIDX_RELEASE,
		EIDX_PACKAGES,
		EIDX_SOURCES,
		EIDX_DIFFIDX,
		EIDX_ARCHLXDB,
		EIDX_CYGSETUP,
		EIDX_SUSEREPO,
		EIDX_SUSEOTHER,
		EIDX_DEBSRCLIKE,
		EIDX_TRANSIDX
	};
	struct tIfileAttribs
	{
		bool vfile_ondisk:1, uptodate:1, parseignore:1, hideDlErrors:1, forgiveDlErrors:1,
		alreadyparsed:1, misc:1;
		enumIndexType eIdxType:8;
		const tStrDeq *bros;

		tIfileAttribs() : vfile_ondisk(false), uptodate(false), parseignore(false),
				hideDlErrors(false), forgiveDlErrors(false), alreadyparsed(false),
				misc(false), eIdxType(EIDX_UNSUPPORTED), bros(NULL)
		{};
	};

	typedef MYSTD::map<mstring,tIfileAttribs> tS2IDX;
	tS2IDX m_indexFilesRel;
	// helpers to keep the code cleaner and more readable
	const tIfileAttribs &GetFlags(cmstring &sPathRel) const;
	tIfileAttribs &SetFlags(cmstring &sPathRel);

	void SetCommonUserFlags(cmstring &cmd);

	void UpdateIndexFiles();
	void _BusyDisplayLogs();
	void _Usermsg(mstring m);
	void AddIFileCandidate(const mstring &sFileRel);

	// NOOP, implemented here for convinience
	bool ProcessOthers(const mstring &sPath, const struct stat &);
	bool ProcessDirAfter(const mstring &sPath, const struct stat &);

	/*!
	 * As the name saids, processes all index files and calls a callback
	 * function maintenence::_HandlePkgEntry on each entry.
	 * 
	 * If a string set object is passed then a little optimization might be 
	 * enabled internally, which avoid repeated processing of a file when another
	 * one with the same contents was already processed. This is only applicable 
	 * having strict path checking disabled, though.
	 *   
	 * */

	void _ProcessSeenIndexFiles();
	bool DownloadIdxFile(const mstring & sFilePathRel, mstring &sErrOut);
	bool DownloadSimple(const mstring & sFilePathRel);
	/*
	void _InjectBz2ed(const mstring & sTargetBasePath, const mstring & sSrcTmpPath,
			const tStrSet & goodChecklist);
			*/

	// internal helper variables
	time_t m_nTimeNow;

	bool m_bErrAbort, m_bVerbose, m_bForceDownload;
	bool m_bScanInternals, m_bByPath, m_bByChecksum, m_bSkipHeaderChecks;
	bool m_bNeedsStrictPathsHere, m_bTruncateDamaged;
	int m_nErrorCount;

	enumIndexType GuessIndexType(const mstring &sPath);

	unsigned int m_nProgIdx, m_nProgTell;

	void TellCount(uint nCount, off_t nSize);

	bool ParseAndProcessIndexFile(ifileprocessor &output_receiver,
			const mstring &sPath, enumIndexType idxType);

private:
	virtual void UpdateFingerprint(const mstring &sPathRel, off_t nOverrideSize,
			uint8_t *pOverrideSha1, uint8_t *pOverrideMd5) =0;
	bool Propagate(const string &donor, tContId2eqClass::iterator eqClassIter);
	void InstallBz2edPatchResult(tContId2eqClass::iterator eqClassIter);
	tCacheMan(const tCacheMan&);
	tCacheMan& operator=(const tCacheMan&);
	bool PatchFile(const mstring &srcRel, const mstring &patchIdxLocation,
			tPListConstIt pit, tPListConstIt itEnd,
			const tFingerprint *verifData);
	dlcon *m_pDlcon;
	void StartDlder();
	tContId2eqClass m_eqClasses;

	bool GetAndCheckHead(cmstring & sHeadfile, cmstring &sFilePathRel, off_t nWantedSize);
	bool Inject(cmstring &from, cmstring &to);

	bool RefetchFile(tFileItemPtr pFi, cmstring &cacheRefFileRel, mstring &sErr);
	bool RecDownload(tFileItemPtr pFi, const char *pURL, int nTTL=3);
};


static const string suxe[] = { ".bz2", ".gz", ".lzma", ".xz"};
static const string suxeWempty[] = { ".bz2", ".gz", ".lzma", ".xz", ""};
static const string suxeByLhood[] = { "", ".bz2", ".gz", ".lzma", ".xz"};
static const string suxeByCompSize[] = { ".xz", ".lzma", ".bz2", ".gz", ""};

class expiration : public tCacheMan
{
public:
	expiration(int);
	virtual ~expiration();

protected:

	tS2DAT m_trashCandSet;
	set<tFileNdir> m_trashCandHeadSet; // just collect the list of seen head files

	void _RemoveAndStoreStatus(bool bPurgeNow);
	void _LoadTrashMapFromFile(bool bForceInsert);

	// callback implementations
	virtual void Action(const mstring &);
	// for FileHandler
	virtual bool ProcessRegular(const mstring &sPath, const struct stat &);

	// for ifileprocessor
	virtual void HandlePkgEntry(const tRemoteFileInfo &entry, bool bUnpackForCsumming);

	virtual void UpdateFingerprint(const mstring &sPathRel, off_t nOverrideSize,
				uint8_t *pOverrideSha1, uint8_t *pOverrideMd5);

	void _LoadIgnoreList();

	void _PurgeMaintLogs();
};

#endif /*EXPIRATION_H_*/
